import inspect

import traceback
from flask import request

from wedeliver_core.helpers.get_embedded_function import (
    get_embedded_function,
)


def validate_parameters(function, app=None):
    """
    This function is used to validate the parameters that are sent to the API
    :param function: The function that is called
    :return:
    """
    # Get the parameters that are required for the function
    try:
        function = get_embedded_function(function)
        if not function:
            return True

        if not request:
            return True
        required_parameters = inspect.getfullargspec(function)

        if (
                not request.original_form
                and not request.original_args
                and len(required_parameters.args) != 0
        ):
            return False
    except Exception:
        app.logger.error(
            "empty_parameters_validation: {0}".format(traceback.format_exc())
        )
        return True
    return True



class AppNoRowsFound(Exception):
    code = 404


class AppException(Exception):
    message = "Error"
    code = 500
    custom_exception = True
    silent = False


class AppSilentException(AppException):
    silent = True


class AppNotSilentException(AppException):
    silent = False


class AppValidationError(AppSilentException):
    code = 400
    message = "Validation Error"


class AppMicroFetcherError(AppSilentException):
    code = 400
    message = "Fetcher Service Error"


class AppThirdPartyError(AppNotSilentException):
    code = 400
    message = "Third Party Call Error"


class AppFetchServiceDataError(AppNotSilentException):
    code = 400
    message = "Error while fetch relational data from service"


class AppNotFoundError(AppSilentException):
    code = 404
    message = "Not found"


class AppMissingAuthError(AppSilentException):
    code = 401
    message = "You are unauthenticated"



import flask_sqlalchemy

from wedeliver_core import WeDeliverCore
from wedeliver_core.helpers.exceptions import AppMicroFetcherError, AppFetchServiceDataError
import requests
import json

from wedeliver_core.helpers.get_obj_value import get_obj_value


class MicroFetcher(object):
    base_data = None
    app = None
    service_name = None
    service_url = None
    fields = None
    table_name = None
    column_name = None
    compair_operator = None
    column_values = None
    output_key = None
    lookup_key = None
    module_name = None
    function_params = None

    def __init__(self, service_name):
        self.app = WeDeliverCore.getApp()
        self.service_url = self.app.config.get(service_name)
        self.service_name = service_name
        if not self.service_url:
            raise AppMicroFetcherError('Service {} not defined on Env'.format(self.service_url))

    def join(self, base_data, output_key=None):
        self.base_data = base_data

        if output_key:
            output_key = output_key.split('as ')[1]

        self.output_key = "{}".format(self.service_name.split('_')[0].lower()) if not output_key else output_key
        return self

    def select(self, *args):
        self.fields = list(args)
        return self

    def filter(self, *args):
        against = args[0].split('.')
        self.compair_operator = args[1]
        self.lookup_key = args[2]
        self.column_values = set()
        if isinstance(self.base_data, dict):
            self.column_values.add(get_obj_value(self.base_data, self.lookup_key))
        else:
            if isinstance(self.base_data, flask_sqlalchemy.Pagination):
                data = self.base_data.items
            else:
                data = self.base_data

            for row in data:
                self.column_values.add(get_obj_value(row, self.lookup_key))

        if not len(self.column_values):
            return self
            # raise AppMicroFetcherError('Lookup key {} not found'.format(self.lookup_key))

        self.column_values = list(self.column_values)
        self.column_values = self.column_values[0] if len(self.column_values) == 1 else self.column_values

        if self.compair_operator not in ('=', 'IN'):
            raise AppMicroFetcherError('Only == currently supported')

        self.compair_operator = 'IN' if isinstance(self.column_values, list) else self.compair_operator

        self.table_name = against[0]
        self.column_name = against[1]

        self.fields.append(self.column_name)
        return self

    def fetch(self):
        if self.column_values or self.module_name:
            return self._call_api()
        else:
            return self.base_data

    def with_params(self, **kwargs):
        self.function_params = kwargs
        return self

    def from_function(self, module_name):
        self.module_name = module_name
        return self

    def _call_api(self):

        url = "{}/fetch_relational_data".format(self.service_url)

        payload_dict = {
            "fields": self.fields,
            "table_name": self.table_name,
            "column_name": self.column_name,
            "compair_operator": self.compair_operator,
            "column_values": self.column_values,
        }

        if self.module_name:
            payload_dict.update(
                functions=[
                    dict(
                        name=self.module_name,
                        fields=self.fields,
                        params=self.function_params if self.function_params else dict()
                    )
                ]
            )

        payload = json.dumps(payload_dict)
        headers = {
            'country_code': 'sa',
            'Content-Type': 'application/json',
        }

        response = requests.request("POST", url, headers=headers, data=payload)

        if response.status_code != 200:
            self.app.logger.error(response.text)
            raise AppFetchServiceDataError()

        result = response.json()
        # in case there is no match from relations data
        # if not len(result):
        #     for rel_value in _relations_values:
        #         fetched_object = dict()
        #         fetched_object['id'] = rel_value
        #         for field in fields:
        #             if not fetched_object.get(field):
        #                 fetched_object[field] = None
        #         result.append(fetched_object)
        if self.base_data:
            return self._map_base(result)

        return result

    def _map_base(self, result):
        if isinstance(self.base_data, dict):
            for rd in result:
                if self.base_data.get(self.lookup_key) == rd.get(self.column_name):
                    try:
                        setattr(self.base_data, self.output_key, rd)
                    except AttributeError:
                        self.base_data[self.output_key] = rd
        else:
            if isinstance(self.base_data, flask_sqlalchemy.Pagination):
                data = self.base_data.items
            else:
                data = self.base_data

            for row in data:
                for rd in result:
                    if get_obj_value(row, self.lookup_key) == rd.get(self.column_name):
                        try:
                            setattr(row, self.output_key, rd)
                        except AttributeError:
                            row[self.output_key] = rd

            if isinstance(self.base_data, flask_sqlalchemy.Pagination):
                self.base_data.items = data
            else:
                self.base_data = data

        return self.base_data



from datetime import datetime, timedelta
import pytz

from flask import request

DATABASE_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
DATABASE_TIME_FORMAT = "%H:%M:%S"


def get_time_zone():
    time_zone = "UTC"
    if request:
        if request.headers.get("country_code"):
            time_zone = dict(sa="Asia/Riyadh", ps="Asia/Hebron").get(
                request.headers.get("country_code"), "UTC"
            )
    return time_zone


def from_utc_to_datetime_zone(date_time, time_zone=None):
    if isinstance(date_time, str):
        date_time = date_time.replace("T", " ").replace("Z", "")
    # Get the default Time Zone if not sent in parameters
    # Time Zone in KSA for example should be: UTC
    if not date_time:
        return date_time

    if len(str(date_time)) < 12:
        return date_time
    if time_zone is None:
        time_zone = get_time_zone()
    # Check if there is a set timezone, if not, return the datetime as it is
    if not time_zone:
        # app.logger.error(format_exception(traceback.format_exc()))
        return date_time

    # Check the time_zone if it is correct. If it is not defined, return the same date_time without conversion
    try:
        time_zone = pytz.timezone(time_zone)
    except Exception:
        # app.logger.error(format_exception(traceback.format_exc()))
        return date_time
    # Convert the date time to datetime object if it is sent as string
    if isinstance(date_time, str) or isinstance(date_time, timedelta):
        date_time = datetime.strptime(str(date_time)[:19], DATABASE_DATE_FORMAT[:17])
    date_time = pytz.utc.localize(date_time)

    # This method is not working, makes some extra minutes, do not use it!!!!
    # date_time = date_time.replace(tzinfo=pytz.utc)
    # If after checking the date time is not an object valid for conversion, return the object as it is
    if not isinstance(date_time, datetime):
        # app.logger.error(format_exception(traceback.format_exc()))
        return date_time

    # Convert the date_time to the target timezone
    # If the conversion failed, we should return the object as it is
    try:
        date_time = (date_time.astimezone(time_zone)).strftime(DATABASE_DATE_FORMAT)
    except Exception:
        pass
        # app.logger.error(format_exception(traceback.format_exc()))

    return date_time


def from_datetime_zone_to_utc(date_time, time_zone=None, only_date=False, to_string=True):
    if isinstance(date_time, str):
        date_time = date_time.replace("T", " ").replace("Z", "")
        if len(str(date_time)) < 12:
            return date_time
    # Get the default Time Zone if not sent in parameters
    # Time Zone in KSA for example should be: UTC

    if time_zone is None:
        time_zone = get_time_zone()
    # Check if there is a set timezone, if not, return the datetime as it is
    if not time_zone:
        # app.logger.error(format_exception(traceback.format_exc()))
        return date_time

    # Check the time_zone if it is correct. If it is not defined,
    # return the same date_time without conversion
    try:
        time_zone = pytz.timezone(time_zone)
    except Exception:
        # app.logger.error(format_exception(traceback.format_exc()))
        return date_time
    # Convert the date time to datetime object if it is sent as string
    if isinstance(date_time, str) or isinstance(date_time, timedelta):
        date_time = datetime.strptime(str(date_time)[:19], DATABASE_DATE_FORMAT[:17])

        date_time = time_zone.localize(date_time)
    # This method is not working, makes somoe extra minutes, do not use it!!!!
    # date_time = date_time.replace(tzinfo=time_zone)

    # If after checking the date time is not an object valid for conversion, return the object as it is
    if not isinstance(date_time, datetime):
        # app.logger.error(format_exception(traceback.format_exc()))
        return date_time

    # Convert the date_time to the target timezone
    # If the conversion failed, we should return the object as it is
    try:
        date_time = (date_time.astimezone(pytz.timezone(pytz.utc.zone)))
        if only_date:
            date_time = datetime(date_time.year, date_time.month, date_time.day)

        if to_string:
            date_time = date_time.strftime(
                DATABASE_DATE_FORMAT
            )
    except Exception:
        pass
        # app.logger.error(format_exception(traceback.format_exc()))
    return date_time



import platform
from datetime import datetime

from .kafka_producer import Producer
from .. import WeDeliverCore

INTERNAL_NOTIFICATION_MESSAGE = "internal_notification_message"


def send_critical_error(message, channel=None):
    channel = channel or "critical-errors"
    send_notification_message(
        channel=channel,
        title="critical",
        color="#df0000",
        message=message,
        emoji=":pleading_face:"
    )


def send_notification_message(
        message, channel="logs", title="Log", color="#32a4a7", emoji=":dizzy_face:"
):
    app = WeDeliverCore.getApp()
    channel = "eng-{0}-{1}".format(
        app.config.get("FLASK_ENV")
        if app.config.get("FLASK_ENV") == "production"
        else "development",
        channel,
    )

    message = "Node: {0}\nEnv:{3}\nDate: {1}\n{2}".format(
        platform.node(), datetime.now(), message, str(app.config.get("FLASK_ENV"))
    )
    print(channel)
    print(message)
    data = {
        "notification_method": "slack",
        "payload": {
            "channel": channel,
            "title": "New {0} in {1} Service".format(
                title, str(app.config.get("SERVICE_NAME"))
            ),
            "text": message,
            "color": color,
            "icon_emoji": emoji,
        },
    }
    app.logger.debug(data)
    Producer().send_topic(topic=INTERNAL_NOTIFICATION_MESSAGE, datajson=data)



from sqlalchemy import text


def sql(query, db):
    result = db.engine.execute(text(query))
    result_list = list()
    # app.logger.debug(result)
    for rowproxy in result:
        temp = dict()
        # rowproxy.items() returns an array like [(key0, value0), (key1, value1)]
        for column, value in rowproxy.items():
            # build up the dictionary
            temp = {**temp, **{column: value}}
        result_list.append(temp)

    return result_list



import inspect
from datetime import datetime
from sys import platform
from wedeliver_core import WeDeliverCore
from flask import request


def format_exception(exception, user=None):
    app = WeDeliverCore.getApp()
    user = user or dict()
    try:
        caller_frame = request.path
    except Exception:
        try:
            caller_frame = inspect.currentframe()
            caller_frame = inspect.getouterframes(caller_frame, 2)[1][3]
        except Exception:
            caller_frame = "Unknown"

    try:
        ip = request.remote_addr
    except Exception:
        ip = "Unknown"
    try:
        args = str(request.args)[:1000]
    except Exception:
        args = "Unknown"
    try:
        form = str(request.form)[:1000]
    except Exception:
        form = "Unknown"
    try:
        user_id = "{0} - {1} - {2}".format(
            user.get("email"),
            user.get("mobile"),
            user.get("user_id"),
        )
    except Exception:
        user_id = "Unknown"
    try:
        role = user.get("role")
    except Exception:
        role = "Unknown"
    try:
        node = platform if type(platform) is str else str(platform.node())
    except Exception:
        node = "Unknown"

    return """Node: {node}
Time: {_time}
Env: {env}
IP: {ip}
API: {call}
User: {user}
Role: {role}
Args: {args}
Form: {form}
Error: {error}""".format(
        env=app.config.get("FLASK_ENV"),
        call=caller_frame,
        user=user_id,
        role=role,
        node=node,
        _time=str(datetime.now()),
        ip=ip,
        args=args,
        form=form,
        error=exception,
    )



import traceback
from functools import wraps
from sqlalchemy.exc import SQLAlchemyError

from wedeliver_core import WeDeliverCore
from wedeliver_core.helpers.auth import Auth
from wedeliver_core.helpers.format_exception import format_exception
from wedeliver_core.helpers.notification_center import (
    send_critical_error,
)
from wedeliver_core.helpers.validate_parameters import (
    validate_parameters,
)


def handle_exceptions(func):
    app = WeDeliverCore.getApp()

    @wraps(func)
    def inner_function(*args, **kws):
        try:
            result = func(*args, **kws)
            return result

        except Exception as e:
            # If the error related to the Database, then close the database session if it is open
            if isinstance(e, SQLAlchemyError):
                try:
                    db = app.extensions['sqlalchemy'].db
                    db.session.close()
                except Exception:
                    pass

            notification_channel = None
            if func is not None:
                if not validate_parameters(function=func):
                    notification_channel = "empty-parameters"

            if hasattr(e, "custom_exception"):
                public_message = e.args[0] if e.args else e.message if hasattr(e, 'message') else 'Unknown'
                status_code = e.code
            else:
                public_message = "Unhandled Exception"
                status_code = 500
                message = format_exception(
                    exception=traceback.format_exc(), user=Auth.get_user()
                )
                app.logger.error(message)
                send_critical_error(message=message, channel=notification_channel)

            return public_message, status_code

    return inner_function



from setuptools import setup, find_namespace_packages

# reading long description from file
# with open('DESCRIPTION.txt') as file:
#     long_description = file.read()


# specify requirements of your package here
REQUIREMENTS = [
    "SQLAlchemy",
    "Flask-SQLAlchemy",
    "flask-marshmallow",
    "confluent-kafka",
    "requests",
    "python-dotenv",
]

# some more details
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Internet',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
]

setup(
    name='wedeliver_core',
    version='0.1.0',
    description='weDeliverCore package',
    long_description="""# Markdown supported!\n\n* weDeliverCore\n* List of features\n""",
    long_description_content_type='text/markdown',
    url='https://www.wedeliverapp.com/',
    author='Eyad Farra',
    author_email='info@wedeliverapp.com',
    license='MIT',
    packages=find_namespace_packages(include=['wedeliver_core', 'wedeliver_core.*']),  # ['wedeliver_core'],
    classifiers=CLASSIFIERS,
    install_requires=REQUIREMENTS
)
